--[[---------------------------------------
	VARIABLE CACHING
-----------------------------------------]]

-- Lua Functions
local floor = math.floor
local pairs, select, tonumber, type, unpack = pairs, select, tonumber, type, unpack
local find, gsub, lower, sub = string.find, gsub, string.lower, string.sub
-- WoW API Functions / Variables
local CreateFont, CreateFrame = CreateFont, CreateFrame
local GetAddOnMetadata = GetAddOnMetadata
local GetSpellInfo = GetSpellInfo
local GetSpecialization, GetSpecializationInfo = GetSpecialization, GetSpecializationInfo
local IsShiftKeyDown, IsControlKeyDown = IsShiftKeyDown, IsControlKeyDown
local UnitClass = UnitClass

local SCA, Ace3, LSM, GUI = unpack(select(2,...))
Ace3.version = GetAddOnMetadata(..., 'Version')

--[[---------------------------------------
	CONSTANTS
-----------------------------------------]]
local SOUND_SOURCE_OPTIONS = {
	["packs"] = "SCA Audio Packs",
	["shared"] = "SharedMedia",
	["custom"] = "Custom Sounds",
}

local SOUND_CHANNEL_OPTIONS = {
	["Master"] = MASTER,
	["SFX"] = SOUND,
	["Music"] = MUSIC_VOLUME,
	["Ambience"] = AMBIENCE_VOLUME,
	["Dialog"] = DIALOG_VOLUME,
}

local SOUND_CUSTOM_TYPES = {
	["file"] = "Single Audio File",
	["pack"] = "Custom Audio Pack",
}

local SOUND_CRIT_TRIGGERS = {
	['crit'] = "Crits Only",
	['nocrit'] = "Non-crits Only",
	['always'] = "Always On",
}

local SOUND_EVENTS = {
	['onCast'] = "On Cast Start",
	['onSuccess'] = "On Cast Success",
	['onFail'] = "On Cast Fail",
	['onDamage'] = "On Damage Dealt",
	['onHeal'] = "On Heal Dealt",
}

local DEFENSE_EVENTS = {
	['absorb'] = COMBAT_TEXT_ABSORB,
	['block'] = COMBAT_TEXT_BLOCK,
	['dodge'] = COMBAT_TEXT_DODGE,
	['parry'] = COMBAT_TEXT_PARRY,
}

local MOVEMENT_ACTIONS = {
	['onStart'] = "On Action Start",
	['onStop'] = "On Action Stop",
}


local function GetSoundInfo(db,event,source,isSeparate)
	local msg = ''
	
	if (isSeparate) then
		if (type(isSeparate) == "string") then
			msg = msg..isSeparate..": "
		elseif (source == "packs" or source == "shared" or source == "custom") then
			msg = msg..gsub(SOUND_EVENTS[event],"On ","")..": "
		end
	end
		
	if (source == "packs") then
		msg = msg.."|cffffd200SCA \"|r|cFF94FF5F"..db[event].packs.text.."|r|cffffd200\" Pack|r"
		
		if (event == "onDamage") then
			msg = msg.." |cFFFF0000|| |cFF73FFF0"..SOUND_CRIT_TRIGGERS[db[event].options.trigger].."|r"
		end
		
		if (event == "onHeal") then
			msg = msg.." |cFFFF0000|| |cFF73FFF0"..SOUND_CRIT_TRIGGERS[db[event].options.trigger].."|r"
		end
		
		if (db[event].packs.isRandom.enabled) then
			msg = msg.." |cFFFF0000|| |cFFEAFF73Random|r"
		else
			msg = msg.." |cFFFF0000|| |cFFEAFF73Sequential|r"
		end
	elseif (source == "shared") then
		msg = msg.."|cffffd200SharedMedia|r |cFFFF0000|| |cFF94FF5F"..db[event].shared.value.."|r"

		if (event == "onDamage") then
			msg = msg.." |cFFFF0000|| |cFF73FFF0"..SOUND_CRIT_TRIGGERS[db[event].options.trigger].."|r"
		end
		
		if (event == "onHeal") then
			msg = msg.." |cFFFF0000|| |cFF73FFF0"..SOUND_CRIT_TRIGGERS[db[event].options.trigger].."|r"
		end
	elseif (source == "custom") then
		msg = msg.."|cffffd200"..SOUND_CUSTOM_TYPES[db[event].custom.customType].."|r |cFFFF0000|| |cFF94FF5F"..db[event].custom.path.."|r"
		
		if (event == "onDamage") then
			msg = msg.." |cFFFF0000|| |cFF73FFF0"..SOUND_CRIT_TRIGGERS[db[event].options.trigger].."|r"
		end
		
		if (event == "onHeal") then
			msg = msg.." |cFFFF0000|| |cFF73FFF0"..SOUND_CRIT_TRIGGERS[db[event].options.trigger].."|r"
		end
		
		if (db[event].custom.customType == "pack") then
			if (tonumber(db[event].custom.max) > 1) then
				msg = msg.." |cFFFF0000|||cFFEC73FF"..db[event].custom.max.." files|r"
			end
		
			if (db[event].custom.isRandom.enabled) then
				msg = msg.." |cFFFF0000|| |cFFEAFF73Random|r"
			else
				msg = msg.." |cFFFF0000|| |cFFEAFF73Sequential|r"
			end
		end
	end

	return msg
end

function Ace3:UpdateSeparateSoundMessage(soundTbl,options,isSpecific,bypass)
	if (not _G["SCA_temp_tbl"]) then
		_G["SCA_temp_tbl"] = options
	end

	local messages = options.args.messageBoard.args.messages
	
	if (soundTbl.isSeparate or bypass) then
		local newMsg = ''
		local msgCache = messages.name
		
		if (isSpecific) then
			local ctr = 1
			for k,v in pairs(soundTbl.events) do
				if (v and Ace3:CheckSpell(k,soundTbl.validSpell)) then
					if (ctr > 1) then
						newMsg = newMsg.."\n"
					end

					newMsg = newMsg..gsub(SOUND_EVENTS[k],"On ","")..": "
					if (soundTbl[k].validAudio) then
						newMsg = newMsg..GetSoundInfo(soundTbl,k,soundTbl[k].options.source)
					else
						newMsg = newMsg.."|cFFFF5F99Not yet configured!|r"
					end
					ctr = ctr + 1
				end
			end
		else
			
			local ctr = 1
			for k,v in pairs(soundTbl.events) do
				if (v) then
					if (ctr > 1) then
						newMsg = newMsg.."\n"
					end

					newMsg = newMsg..gsub(soundTbl[k].label,"On ","")..": "
					if (soundTbl[k].validAudio) then
						newMsg = newMsg..GetSoundInfo(soundTbl,k,soundTbl[k].options.source)
					else
						newMsg = newMsg.."|cFFFF5F99Not yet configured!|r"
					end
					ctr = ctr + 1
				end
			end
		end
		
		if (soundTbl.messages) then
			soundTbl.messages.separateSounds = newMsg
		end

		messages.name = newMsg
	else
		if (soundTbl.messages) then
			soundTbl.messages.separateSounds = ''
		end
		messages.name = ' '
	end
end

local function InitMoveOptions(db)
	for k,_ in pairs(db) do
		if (k ~= "enabled") then
			db[k].numEvents = 0
			for j,v in pairs(db[k].events) do
				if (v) then
					db[k].numEvents = db[k].numEvents + 1
				end
			end
		end
	end
end

local function InitSpellOptions(db,options)
	db.numEvents = 0
	db.tutorial.enabled = false
	db.tutorial.ids = nil
	db.tutorial.ids = {}

	if (db.isSeparate) then
		Ace3:UpdateSeparateSoundMessage(db,options)
	end

	for k,v in pairs(db.events) do
		if (v) then
			db.numEvents = db.numEvents + 1
		end
	end
end

local function RelocateTutorialPanel(db,panel,ignoreSeparate)
	local tutorial = db.tutorial
	local frame = NPE_TutorialPointerFrame.InUseFrames[tutorial.ids[panel]]
	
	if (frame) then
		local point,parent,relativePoint = frame:GetPoint()
		local newY = 0
		local oldX = tutorial.ofs[panel].x
		local oldY = tutorial.ofs[panel].y

		frame:ClearAllPoints()
		
		if (db.isSeparate and SCA.temp.errors[db.dbName] and not ignoreSeparate) then
			if (panel == 'customType') then
			end
			newY = oldY - 122
		elseif (db.isSeparate and not ignoreSeparate) then
			if (panel == 'customType') then
			end
			newY = oldY - 66
		elseif (SCA.temp.errors[db.dbName]) then
			if (panel == 'customType') then
			end
			newY = oldY - 56
		else
			if (panel == 'customType') then
			end
			newY = oldY
		end
		
		frame:SetPoint(point,parent,relativePoint,oldX,newY)
	end
end

local function DestroyTutorialPanel(db,panel)
	local tutorial = db.tutorial
	NPE_TutorialPointerFrame:Hide(tutorial.ids[panel])
	tutorial.ids[panel] = nil
	tutorial.frameCount = tutorial.frameCount - 1
end

function Ace3:UpdateTutorial(db,trigger)
	local tutorial = db.tutorial
	
	if (tutorial.enabled) then
		if (not IsAddOnLoaded("Blizzard_Tutorial")) then
			LoadAddOn("Blizzard_Tutorial")
		end
		
		if (db.validSpell == false) then
			tutorial.ids.spellName = NPE_TutorialPointerFrame:Show("In order to access the sound controls, you first must input a valid spell name or ID.\n\n|cFF00FF00Read the textbox's tooltip for more details.|r","DOWN",InterfaceOptionsFrame,331,-150,"TOPLEFT")
			tutorial.frameCount = tutorial.frameCount + 1
		else
			if (db.validSpell and db.selected == "empty" and (trigger ~= "toggle" or trigger == "toggle")) then
				db.selected = "onSuccess"
			end
			
			if (tutorial.ids.spellName) then
				DestroyTutorialPanel(db,'spellName')
			end

			if (not tutorial.ids.spellEvents and not tutorial.flags.spellEvents) then
				tutorial.ids.spellEvents = NPE_TutorialPointerFrame:Show("Toggling these \"|cFFFFFFFFSpell Events|r\" will set which events will trigger sounds.","DOWN",InterfaceOptionsFrame,tutorial.ofs.spellEvents.x,tutorial.ofs.spellEvents.y,"TOPLEFT")
				tutorial.frameCount = tutorial.frameCount + 1
			else
				if (tutorial.flags.spellEvents) then
					DestroyTutorialPanel(db,'spellEvents')
				end
			end
			
			if (not tutorial.ids.source and not tutorial.flags.source and db[db.selected].options.source == "packs") then
				tutorial.ids.source = NPE_TutorialPointerFrame:Show("Using this dropdown allows you to select the source of sounds you want to choose from.","RIGHT",InterfaceOptionsFrame,tutorial.ofs.source.x,tutorial.ofs.source.y,"TOPLEFT")
				tutorial.frameCount = tutorial.frameCount + 1
				
				if (db.isSeparate or SCA.temp.errors[db.dbName]) then
					RelocateTutorialPanel(db,'source')
				end
			else
				if (tutorial.flags.source) then
					DestroyTutorialPanel(db,'source')
				end
				
				RelocateTutorialPanel(db,'source')
			end
			
			if (not tutorial.ids.isSeparate and not db.isSeparate and not tutorial.flags.isSeparate) then 
				tutorial.ids.isSeparate = NPE_TutorialPointerFrame:Show("Toggling this will allow the ability to separate the spell events; allowing you to add a specific sound for each event.\n\n|cFFFF5F99Having more than one spell event selected is required.|r","LEFT",InterfaceOptionsFrame,tutorial.ofs.isSeparate.x,tutorial.ofs.isSeparate.y,"TOPLEFT")
				tutorial.frameCount = tutorial.frameCount + 1
			else
				if (db.isSeparate or tutorial.flags.isSeparate) then
					DestroyTutorialPanel(db,'isSeparate')
				end
				
				RelocateTutorialPanel(db,'isSeparate')
			end
			
			if (db.isSeparate) then
				if (not tutorial.ids.spellEvent and not tutorial.flags.spellEvent) then
					tutorial.ids.spellEvent = NPE_TutorialPointerFrame:Show("Use this dropdown to select the specific spell events to assign sounds.","DOWN",InterfaceOptionsFrame,tutorial.ofs.spellEvent.x,tutorial.ofs.spellEvent.y,"TOPLEFT")
					tutorial.frameCount = tutorial.frameCount + 1
				else
					if (tutorial.flags.spellEvent) then
						DestroyTutorialPanel(db,'spellEvent')
					end
					
					RelocateTutorialPanel(db,'spellEvent',true)
				end
			else
				if (tutorial.ids.spellEvent) then
					DestroyTutorialPanel(db,'spellEvent')
				end
				
				RelocateTutorialPanel(db,'spellEvent',true)
			end
			
			if (db[db.selected].options.source == "custom") then
				if (not tutorial.ids.customType and not tutorial.flags.customType) then
					tutorial.ids.customType = NPE_TutorialPointerFrame:Show("This dropdown sets up the ability to add a single sound file or a custom sound pack","RIGHT",InterfaceOptionsFrame,tutorial.ofs.customType.x,tutorial.ofs.customType.y,"TOPLEFT")
					tutorial.frameCount = tutorial.frameCount + 1
				
					if (db.isSeparate or SCA.temp.errors[db.dbName]) then
						RelocateTutorialPanel(db,'customType')
					end
				else
					if (tutorial.flags.customType and tutorial.ids.customType) then
						DestroyTutorialPanel(db,'customType')
					end
					
					RelocateTutorialPanel(db,'customType')
				end
				
				if (not tutorial.ids.audioLocationFile) then
					if (db[db.selected].custom.customType == "file" and not tutorial.flags.audioLocationFile) then
						tutorial.ids.audioLocationFile = NPE_TutorialPointerFrame:Show("Set the location of the audio file. The file |cFF00FF00MUST|r be located somewhere within the \"|cFFFFFFFFInterface|r\" folder.\r\r|cFF00FF00Recommended:\nInterface\\audio\\<file>","UP",InterfaceOptionsFrame,tutorial.ofs.audioLocationFile.x,tutorial.ofs.audioLocationFile.y,"TOPLEFT")
						tutorial.frameCount = tutorial.frameCount + 1
				
						if (db.isSeparate or SCA.temp.errors[db.dbName]) then
							RelocateTutorialPanel(db,'audioLocationFile')
						end
					end
				else
					if (tutorial.flags.audioLocationFile or db[db.selected].custom.customType == "pack") then
						DestroyTutorialPanel(db,'audioLocationFile')
					end
					
					RelocateTutorialPanel(db,'audioLocationFile')
				end
				
				if (not tutorial.ids.audioLocationPack) then
					if (db[db.selected].custom.customType == "pack" and not tutorial.flags.audioLocationPack) then
						tutorial.ids.audioLocationPack = NPE_TutorialPointerFrame:Show("Set the location of the audio pack. The pack |cFF00FF00MUST|r be located somewhere within the \"|cFFFFFFFFInterface|r\" folder.\r\r|cFF00FF00Recommended:\nInterface\\audio\\<pack>\\","UP",InterfaceOptionsFrame,tutorial.ofs.audioLocationPack.x,tutorial.ofs.audioLocationPack.y,"TOPLEFT")
						tutorial.frameCount = tutorial.frameCount + 1
				
						if (db.isSeparate or SCA.temp.errors[db.dbName]) then
							RelocateTutorialPanel(db,'audioLocationPack')
						end
					end
				else
					if (tutorial.flags.audioLocationPack or db[db.selected].custom.customType == "file") then
						DestroyTutorialPanel(db,'audioLocationPack')
					end
					
					RelocateTutorialPanel(db,'audioLocationPack')
				end
				
				if (not tutorial.ids.numFiles) then
					if (db[db.selected].custom.customType == "pack" and not tutorial.flags.numFiles) then
						tutorial.ids.numFiles = NPE_TutorialPointerFrame:Show("In order for the pack to be registered to work correctly, you must enter the number of files within the custom pack folder.","UP",InterfaceOptionsFrame,tutorial.ofs.numFiles.x,tutorial.ofs.numFiles.y,"TOPLEFT")
						tutorial.frameCount = tutorial.frameCount + 1
				
						if (db.isSeparate or SCA.temp.errors[db.dbName]) then
							RelocateTutorialPanel(db,'numFiles')
						end
					end
				else
					if (tutorial.flags.numFiles) then
						DestroyTutorialPanel(db,'numFiles')
					end
					
					RelocateTutorialPanel(db,'numFiles')
				end
			else
				if (tutorial.ids.customType) then
					DestroyTutorialPanel(db,'customType')
				end
				
				if (tutorial.ids.audioLocationFile) then
					DestroyTutorialPanel(db,'audioLocationFile')
				end
				
				if (tutorial.ids.audioLocationPack) then
					DestroyTutorialPanel(db,'audioLocationPack')
				end
				
				if (tutorial.ids.numFiles) then
					DestroyTutorialPanel(db,'numFiles')
				end
			end
			
			for k,v in pairs(tutorial.ids) do
				if (tutorial.ids[k]) then
					NPE_TutorialPointerFrame.InUseFrames[tutorial.ids[k]].Content:SetScript("OnMouseDown",function(self,button)
						if (button == "RightButton") then
							DestroyTutorialPanel(db,k)
						end
						
						if (tutorial.frameCount == 0) then
							tutorial.enabled = false
						end
					end)
				end
			end
		end
	else
		for k,_ in pairs(tutorial.ids) do
			DestroyTutorialPanel(db,k)
		end
	end
end

function Ace3:VerifyCustomPath(value,soundInfo)
	if (find(value,".mp3")) then
		value = gsub(value,".mp3",'')
	elseif (find(value,".ogg")) then
		value = gsub(value,".ogg",'')
	end
	
	if (value ~= "Interface\\AUDIO\\") then
		if (sub(value,#value) ~= [[\]] and soundInfo.custom.customType == "pack" and #value > 0) then
			value = value..[[\]]
		elseif (sub(value,#value) == [[\]] and soundInfo.custom.customType == "file") then
			value = sub(value,0,#value - 1)
		end
	end
	
	return value
end

local function ReloadSpellIgnoreList(options,tab)
	local db = Ace3.db.char
	local args = {}
	local sortTable = {}

	for k,_ in pairs(db.audio.spells.ignored) do
		local name = GetSpellInfo(k)
		
		tinsert(sortTable,name)
	end
	
	table.sort(sortTable)
	
	for i=1,#sortTable do
		local _,spellID,name = Ace3:CheckSpell('onSuccess',sortTable[i])
		
		args["spell"..i] = {
			order = i,
			type = "toggle",
			desc = "Spell ID: "..spellID,
			descStyle = "inline",
			name = "|cFFEBC000"..name.."|r",
			get = function() 
				return true
			end,
			set = function(self,value)
				db.audio.spells.ignored[spellID] = nil
				ReloadSpellIgnoreList(options)
			end,
		}
	end

	options.args.ignoreTab.args.ignoreList.args = args
end

local function ReloadSpellAssignList(options,db)
	local args = {}
	local _,className = UnitClass("player")
	local spec = GetSpecialization()
	local sortTable = {}
	
	-- Insert spell names from "Specific" table into local table to be sorted
	for k,v in pairs(db.specific) do
		if (SCA.overloadSpells[className][spec][k]) then
			tinsert(sortTable,SCA.overloadSpells[className][spec][k])
		else
			tinsert(sortTable,SCA.onSuccessSpells[className][spec][k])
		end
	end
	
	table.sort(sortTable)
	
	for i=1,#sortTable do
		local name = sortTable[i]

		if (SCA.spellNames[className][spec][sortTable[i]] or SCA.overloadSpells[className][spec][sortTable[i]]) then
			local specificTbl = db.specific[SCA.spellNames[className][spec][sortTable[i]]]
			if (i == 1) then
				args["help"] = {
					order = 1,
					type = "description",
					name = "EDIT: Shift+Click. DELETE: Ctrl+Click",
					desc = "HELP TIPS",
				}
			end
			args["spell"..i] = {
				order = i+1,
				type = "toggle",
				desc = function()
					local msg = ''
					local ctr = 0
					for k,v in pairs(SOUND_EVENTS) do
						if (specificTbl[k]) then
							if (specificTbl.isSeparate) then
								if (ctr > 0) then
									msg = msg.."\n"
								end
								
								for j,_ in pairs(specificTbl[k]) do
									msg = msg..GetSoundInfo(specificTbl,k,j,specificTbl.isSeparate)
									
									ctr = ctr + 1
								end
							else
								msg = msg..GetSoundInfo(specificTbl,k,specificTbl[k].options.source)

								break
							end
						end
					end
						
					return msg
				end,
				descStyle = "inline",
				name = "|cFFEBC000"..name.."|r",
				get = function() 
					return true
				end,
				set = function(self,value)
					if (IsShiftKeyDown()) then
						local ctr = 0
						local newSound = SCA.temp.newSound
						
						SCA.temp.addSound = true
						SCA.temp.editSound = true
						
						newSound.validSpell = name
						
						newSound.isSeparate = specificTbl.isSeparate
						
						for k,_ in pairs(SOUND_EVENTS) do
							if (specificTbl[k]) then
								newSound[k].validAudio = true
								newSound.events[k] = true
								
								newSound[k].options.source = specificTbl[k].options.source
								newSound[k].options.channel = specificTbl[k].options.channel
								
								if (k == "onDamage" or k == "onHeal") then
									newSound[k].options.trigger = specificTbl[k].options.trigger
									newSound[k].options.threshold = specificTbl[k].options.threshold
								end
								
								if (specificTbl[k].packs) then
									newSound[k].packs.value = specificTbl[k].packs.value
									newSound[k].packs.text = specificTbl[k].packs.text
									newSound[k].packs.path = specificTbl[k].packs.path
									newSound[k].packs.max = specificTbl[k].packs.max
									newSound[k].packs.isRandom.enabled = specificTbl[k].packs.isRandom.enabled
									newSound[k].packs.isRandom.currentFile = specificTbl[k].packs.isRandom.currentFile
								end
								
								if (specificTbl[k].shared) then
									newSound[k].shared.value = specificTbl[k].shared.value
								end
								
								if (specificTbl[k].custom) then
									newSound[k].custom.customType = specificTbl[k].custom.customType
									newSound[k].custom.path = specificTbl[k].custom.path
									newSound[k].custom.max = specificTbl[k].custom.max
									newSound[k].custom.isRandom.enabled = specificTbl[k].custom.isRandom.enabled
									newSound[k].custom.isRandom.currentFile = specificTbl[k].custom.isRandom.currentFile
								end
								ctr = ctr + 1
							end
						end
						
						if (newSound.isSeparate) then
							Ace3:UpdateSeparateSoundMessage(newSound,self.options.args.SpecificTab,true)
						end
						
						newSound.numEvents = ctr
					elseif (IsControlKeyDown()) then
						db.specific[SCA.spellNames[className][spec][sortTable[i]]] = nil
						ReloadSpellAssignList(options,db)
					end
				end,
				width = "full",
			}
		end
	end

	options.args.SpecificTab.args.assignedSpells.args = args
end

function Ace3:ResetDB(db,isResetAll,isNewSound)
	if (isNewSound) then
		SCA.temp.addSound = false
		SCA.temp.editSound = false
	end
	
	if (isResetAll) then
		if (db.validSpell) then
			db.validSpell = false
		end
		if (db.numEvents) then
			db.numEvents = 0
		end
		
		if (db.isSeparate) then
			db.isSeparate = false
		end
		
		if (db.messages) then
			db.messages.separateSounds = ''
			db.messages.notices = ''
		end
		
		if (db.selected) then
			db.selected = 'onCast'
		end

		if (db.tutorial) then
			db.tutorial.enabled = false
			db.tutorial.flags.isSeparate = false
			db.tutorial.flags.spellEvent = false
			db.tutorial.flags.spellEvents = false
			db.tutorial.flags.source = false
			db.tutorial.flags.customType = false
			db.tutorial.flags.numFiles = false
			db.tutorial.flags.audioLocationPack = false
			db.tutorial.flags.audioLocationFile = false
			
			db.tutorial.ids = nil
			db.tutorial.ids = {}
		end
	end
	
	for k,_ in pairs(db.events) do
		if (isResetAll) then
			db.events[k] = false
			
			db[k].options.source = "packs"
			db[k].options.channel = "Master"
			
			if (k == "onDamage" or k == "onHeal") then
				db[k].options.trigger = "crit"
				db[k].options.threshold = 0
			end
		end
		
		if (db[k].validAudio) then
			db[k].validAudio = false
		end
		
		db[k].packs.value = "empty"
		db[k].packs.text = ''
		db[k].packs.path = ''
		db[k].packs.max = 1
		db[k].packs.isRandom.enabled = true
		db[k].packs.isRandom.currentFile = 1
		
		db[k].shared.value = "None"
		
		db[k].custom.path = "Interface\\AUDIO\\"
		db[k].custom.max = 1
		db[k].custom.customType = "file"
		db[k].custom.isRandom.enabled = true
		db[k].custom.isRandom.currentFile = 1
	end
end

local function CheckIgnoreSpell(newIgnore,errors,isValidSpell,spellID,name,errorMsg)
	local spells = Ace3.db.char.audio.spells
	
	if (isValidSpell and not spells.ignored[spellID]) then
		errors.name = " "
		
		newIgnore.validSpell = name
		SCA.temp.errors.ignored = nil
		
		
		if (spells.specific[spellID]) then
			errors.name = name.." |cFFFF9B00has a sound assigned to it. Ignoring it will remove its sound|r"
			SCA.temp.errors.ignored = true
		end
	else
		if (spells.ignored[spellID]) then
			errors.name = "|cFFFF5F99Spell is already ignored|r"
		else
			errors.name = errorMsg
		end
		
		newIgnore.validSpell = ''
		SCA.temp.errors.ignored = true
	end
end

local function CheckSpecificSpell(newSound,errors,isValidSpell,spellID,name,errorMsg)
	local spells = Ace3.db.char.audio.spells
	
	if (isValidSpell) then
		newSound.numEvents = 0
		newSound.validSpell = name
		
		if (spells.specific[spellID]) then
			errors.name = name.." |cFFFF9B00's sound will be updated!|r"
			SCA.temp.errors.newSound = true
		elseif (spells.ignored[spellID]) then
			errors.name = name.." |cFFFF9B00is being ignored and will be removed from the ignored list|r"
			SCA.temp.errors.newSound = true
		else
			errors.name = " "
			SCA.temp.errors.newSound = nil
		end
		
		for k,_ in pairs(newSound.events) do
			if (Ace3:CheckSpell(k,name)) then
				newSound.numEvents = newSound.numEvents + 1
				newSound.events[k] = true
			else
				newSound.events[k] = false
			end
		end
	else
		errors.name = errorMsg
		SCA.temp.errors.newSound = true
		newSound.validSpell = false
	end
end

local function GroupBuilder(db,width,eventOnly,toggleOnly,bypass)
	local args = {}
	local order = 1

	if (not eventOnly or toggleOnly) then
		local root
		if (bypass) then
			root = db[bypass]
		else
			root = db
		end
		
		args["allToggle"] = {
			order = 1,
			type = "toggle",
			name = "Enabled",
			desc = '',
			get = function()
				return root.enabled
			end,
			set = function(self,value)
				root.enabled = value
			end,
		}
		args["spacing"] = {
			order = 2,
			type = "description",
			name = " ",
			width = "double",
		}
		order = order + 2
	end
	
	if (not toggleOnly) then
		local events,root
		if (bypass) then
			events = db[bypass].events
			root = db[bypass]
		else
			events = db.events
			root = db
		end
		
		for k,v in pairs(events) do
			args[k.."Toggle"] = {
				order = order,
				type = "toggle",
				name = root[k].label,
				desc = '',
				disabled = function()
					if (not eventOnly and not bypass) then
						if (db.enabled) then
							return false
						else
							return true
						end
					elseif (bypass) then
						if (not db.enabled) then
							return true
						else
							return not root.enabled
						end
					end
				end,
				get = function()
					return events[k]
				end,
				set = function(self,value)
					events[k] = value
				end,
				width = width,
			}
			order = order + 1
		end
	end
	
	return args
end

local function SubGroupBuilder(db)
	local args = {}
	
	args["GlobalToggle"] = {
		order = 1,
		type = "toggle",
		name = "Enabled",
		desc = '',
		get = function()
			return db.enabled
		end,
		set = function(self,value)
			db.enabled = value
		end,
	}
	
	local ctr = 1
	for k,_ in pairs(db) do
		if (type(db[k]) == "table") then
			args[k] = {
				name = db[k].label,
				order = db[k].order,
				type = "group",
				disabled = function()
					if (db.enabled) then
						return false
					else
						return true
					end
				end,
				args = GroupBuilder(db,nil,nil,nil,k),
			}
			ctr = ctr + 1
		end
	end
	
	return args
end

local toggle_options

local function GetToggleOptions()
	SCA.temp.selectedMenuItem = 1
	if (not toggle_options) then
		local db = Ace3.db.char
		
		local damage = db.audio.spells.damage
		local heals = db.audio.spells.heals
		local specific = db.audio.spells.specific
		
		local defense = db.audio.defense
		local special = db.audio.special
		
		local chat = db.audio.chat
		local move = db.audio.move
		
		toggle_options = {
			type = "group",
			order = 1,
			name = "",
			desc = '',
			guiInline = true,
			args = {
				DamageSpells = {
					name = "Damage Spells",
					order = 1,
					type = "group",
					args = GroupBuilder(damage,"normal"),
				},
				HealingSpells = {
					name = "Healing Spells",
					order = 2,
					type = "group",
					args = GroupBuilder(heals,"normal"),
				},
				SpecificSpells = {
					name = "Specific Spells",
					order = 3,
					type = "group",
					args = GroupBuilder(specific,nil,nil,true),
				},
				Defense = {
					name = "Defense Sounds",
					order = 4,
					type = "group",
					args = GroupBuilder(defense,"half",true),
				},
				Special = {
					name = "Special Sounds",
					order = 5,
					type = "group",
					args = SubGroupBuilder(special),
				},
				Chat = {
					name = "Chat Sounds",
					order = 6,
					type = "group",
					args = SubGroupBuilder(chat),
				},
				Movement = {
					name = "Movement Sounds",
					order = 6,
					type = "group",
					args = SubGroupBuilder(move),
				},
			},
		}
	end
	
	return toggle_options
end

local function SpellTabBuilder(db,title,event)
	local tab = {
		name = title,
		order = 1,
		type = "group",
		args = {
			disabledError = Ace3:DisabledError(db),
			SpellConfig = {
				name = title.." Events",
				order = 1,
				type = "group",
				guiInline = true,
				args = {
					onCast = Ace3:EventToggle(db,'onCast',4,"half","Start","Toggles whether a sound will occur when starting a spell cast.\n\nCombat Log Event: |cFF00FF00SPELL_CAST_START|r",title.."Tab"),
					onSuccess = Ace3:EventToggle(db,'onSuccess',5,"half","Success","Toggles whether a sound will occur when a spell cast completes.\n\nCombat Log Event: |cFF00FF00SPELL_CAST_SUCCESS|r",title.."Tab"),
					onFail = Ace3:EventToggle(db,'onFail',6,"half","Failed","Toggles whether a sound will occur when a spell cast fails.\n\nAlso triggered by being interrupted.\n\nCombat Log Event: |cFF00FF00SPELL_CAST_FAILED|r",title.."Tab"),
					[event] = Ace3:EventToggle(db,event,8,"half",title,"Toggles whether a sound will occur when starting a spell causes damage.\n\nCombat Log Events: |cFF00FF00SPELL_"..strupper(title).."|r, |cFF00FF00SPELL_PERIODIC_"..strupper(title).."|r",title.."Tab"),
					spacing1 = Ace3:Spacer(9,"half"),
					tutorial = Ace3:TutorialToggle(db,10,"half"),
				},
			},
			messageBoard = Ace3:SoundList(db,2),
			errors = Ace3:Errors(lower(title),3),
			SoundConfig = {
				name = "Audio Config",
				order = 5,
				type = "group",
				disabled = function()
					if (db.numEvents == 0) then
						return true
					else
						return false
					end
				end,
				guiInline = true,
				args = {
					source = Ace3:Source(db,lower(title),SOUND_SOURCE_OPTIONS),
					channel = Ace3:Channel(db,SOUND_CHANNEL_OPTIONS),
					volume = Ace3:Volume(db),
					--[[-----------------------------
						SCA Sound Packs
					-------------------------------]]
					packSelect = Ace3:PackSelect(db,title.."Tab"),
					--[[-----------------------------
						SharedMedia
					-------------------------------]]
					sharedMedia = Ace3:SharedMedia(db,title.."Tab"),
					--[[-----------------------------
						Custom Audio
					-------------------------------]]
					audioSource = Ace3:AudioSource(db,SOUND_CUSTOM_TYPES,title.."Tab"),
					soundEvents = {
						order = 5,
						type = "select",
						name = "Spell Events",
						desc = "Choose whether your custom audio is a file or sound pack",
						disabled = function()
							if (db.isSeparate and db.numEvents > 1) then
								return false
							else
								return true
							end
						end,
						get = function()
							if (db.events.onCast) then
								if (db.selected ~= "onCast" and not db.events.onSuccess and not db.events.onFail and not db.events[event]) then
									db.selected = 'onCast'
								end
							else
								if (db.selected == "onCast") then
									for k,v in pairs(db.events) do
										if (v) then
											db.selected = k
											break
										end
									end
												
									if (db.selected == "onCast") then
										db.selected = "empty"
									end
								end
							end
							
							if (db.events.onSuccess) then
								if (db.selected ~= "onSuccess" and not db.events.onCast and not db.events.onFail and not db.events[event]) then
									db.selected = 'onSuccess'
								end
							else
								if (db.selected == "onSuccess") then
									for k,v in pairs(db.events) do
										if (v) then
											db.selected = k
											break
										end
									end
												
									if (db.selected == "onSuccess") then
										db.selected = "empty"
									end
								end
							end
							
							if (db.events.onFail) then
								if (db.selected ~= "onFail" and not db.events.onCast and not db.events.onSuccess and not db.events[event]) then
									db.selected = 'onFail'
								end
							else
								if (db.selected == "onFail") then
									for k,v in pairs(db.events) do
										if (v) then
											db.selected = k
											break
										end
									end
												
									if (db.selected == "onFail") then
										db.selected = "empty"
									end
								end
							end
							
							if (db.events[event]) then
								if (db.selected ~= event and not db.events.onCast and not db.events.onSuccess and not db.events.onFail) then
									db.selected = event
								end
							else
								if (db.selected == event) then
									for k,v in pairs(db.events) do
										if (v) then
											db.selected = k
											break
										end
									end
												
									if (db.selected == event) then
										db.selected = "empty"
									end
								end
							end
							
							return db.selected
						end,
						set = function(self,value)
							db.selected = value
							
							db.tutorial.flags.spellEvent = true
							Ace3:UpdateTutorial(db)
						end,
						values = function()
							local args = {}
							
							for k,v in pairs(db.events) do
								if (v) then
									args[k] = SOUND_EVENTS[k]
								end
							end

							if (db.numEvents == 0) then
								args["empty"] = "No events"
							end
							
							return args
						end
					},
					toggleSeparate = Ace3:ToggleSeparate(db,title..'Tab'),
					isSCARandom = Ace3:IsSCARandom(db),
					audioLocation = Ace3:AudioLocation(db),
					numFiles = Ace3:NumFiles(db,lower(title),title..'Tab'),
					test = Ace3:TestSound(db,lower(title),title..'Tab'),
					isCustomRandom = Ace3:IsCustomRandom(db),
					reset = Ace3:SoftReset(db,title..'Tab'),
				},
			},
			[title.."TriggerHeader"] = {
				order = 6,
				type = 'group',
				name = title..' Triggers',
				guiInline = true,
				hidden = function()
					if (db.events[event]) then
						return false
					else
						return true
					end
				end,
				args = {
					spells = {
						order = 1,
						type = "toggle",
						name = "Spells",
						desc = '',
						get = function()
							return db[event].options.spells
						end,
						set = function(self,value)
							db[event].options.spells = value
						end,
					},
					[lower(sub(title,0,1)).."ots"] = {
						order = 2,
						type = "toggle",
						name = sub(title,0,1).."oTs",
						desc = '',
						get = function()
							return db[event].options[lower(sub(title,0,1)).."ots"]
						end,
						set = function(self,value)
							db[event].options[lower(sub(title,0,1)).."ots"] = value
						end,
					},
					pets = {
						order = 3,
						type = "toggle",
						name = "Pets",
						desc = "Damage that your pet does",
						hidden = function()
							if (event == "onDamage") then
								return false
							else
								return true
							end
						end,
						get = function()
							return db.onDamage.options.pets
						end,
						set = function(self,value)
							db.onDamage.options.pets = value
						end,
					},
					spacing = {
						order = 4,
						type = "description",
						name = " ",
						hidden = function()
							if (event == "onHeal") then
								return false
							else
								return true
							end
						end,
						width = "normal",
					},
					threshold = Ace3:Threshold(db,event,title,lower(title),title..'Tab'),
					critSelect = Ace3:CritSelect(db,event,SOUND_CRIT_TRIGGERS),
					swing = {
						order = 7,
						type = "toggle",
						name = "Auto-Attacks",
						hidden = function()
							if (event == "onDamage") then
								return false
							else
								return true
							end
						end,
						desc = '',
						get = function()
							return db.onDamage.options.swing
						end,
						set = function(self,value)
							db.onDamage.options.swing = value
						end,
					},
				},
			},
			reset = Ace3:HardReset(db,title.."Tab"),
		},
	}
	
	return tab
end

local spell_options

local function GetSpellOptions()
	SCA.temp.selectedMenuItem = 2
	if (not spell_options) then
		--local db = Ace3.db.char
		local db = Ace3.db.char
		
		local damage = db.audio.spells.damage
		local heals = db.audio.spells.heals
		local specific = db.audio.spells.specific
		local ignored = db.audio.spells.ignored
		
		local newSound = SCA.temp.newSound
		local newIgnore = SCA.temp.newIgnore
		
		spell_options = {
			type = "group",
			childGroups = "tab",
			order = 1,
			name = "",
			desc = '',
			args = {
				DamageTab = SpellTabBuilder(damage,"Damage","onDamage"),
				HealingTab = SpellTabBuilder(heals,"Healing","onHeal"),
				SpecificTab = {
					name = "Specific Spells",
					order = 3,
					type = "group",
					args = {
						disabledError = Ace3:DisabledError(specific),
						AddCancelSound = {
							order = function()
								if (SCA.temp.addSound) then
									return 9
								else
									return 1
								end
							end,
							type = "execute",
							name = function(self)
								if (SCA.temp.addSound) then
									self.options.args.SpecificTab.args.AddCancelSound.width = "half"
								else
									self.options.args.SpecificTab.args.AddCancelSound.width = "normal"
								end
								
								if (SCA.temp.addSound) then
									return "Cancel"
								else
									return "Add Sound"
								end
							end,
							func = function()
								if (SCA.temp.addSound) then
									if (SCA.SpellFrame:IsShown()) then
											SCA.SpellFrame:Hide()
									end
									Ace3:ResetDB(newSound,true,true)
								else
									SCA.temp.addSound = true
								end
							end,
							width = "normal",
						},
						SpellConfig = {
							name = "Spell Config",
							hidden = function()
								return not SCA.temp.addSound
							end,
							order = 2,
							type = "group",
							guiInline = true,
							args = {
								spellInfo = {
									type = "input",
									order = 1,				
									name = "Spell ID / Name",
									desc = "Input spell id, or name, to be assigned a specific sound.\n\nSpell names and IDs that you input are checked against a pre-built table of spell ID and names. If there is a spell that is missing that you feel should be added, please message me via the addon page.\n\nClicking the \"Spells\" button will display a list of spells that currently exist for your current class/spec.",
									disabled = function()
										if (SCA.temp.editSound) then
											return true
										else
											return false
										end
									end,
									get = function(self)
										return newSound.validSpell
									end,
									set = function(self,value)
										local errors = self.options.args.SpecificTab.args.errors.args.messages
										
										CheckSpecificSpell(newSound,errors,Ace3:CheckSpell('onSuccess',value))
										
										Ace3:UpdateTutorial(newSound)
										
										-- If no event is selected ("empty"), we'll need to refresh the config's contents.
										if (newSound.selected == "empty") then
											Ace3:NavigateInterfaceOptions(SCA.temp.selectedMenuItem,true)
										end
									end,
								},
								spells = {
									order = 2,
									type = "execute",
									name = "Spells",
									desc = "List of spells, and spell IDs, that pertain to your class and spec.\n\n|cFFFF9B00Orange Spells|r are spells that are currently on the ignore list.\n\nA |cFF00FF00Green Spell|r is the spell that is currently selected.\n\n|cFF999999Grey Spells|r are spells that already have sounds assigned to them.",
									disabled = function()
										if (SCA.temp.editSound) then
											return true
										else
											return false
										end
									end,
									func = function()
										local _,class = UnitClass("player")
										local spec = GetSpecialization()
										if (SCA.SpellFrame:IsShown()) then
											SCA.SpellFrame:Hide()
										else
											for k,_ in pairs(SCA.spellNames[class][spec]) do
												if (SCA.SpellFrame[k]) then
													if (specific[SCA.spellNames[class][spec][k]]) then
														SCA.SpellFrame[k]:Disable()
														SCA.SpellFrame[k]:SetDisabledFontObject(SCA.SpellFrame.dFont)
													elseif (ignored[SCA.spellNames[class][spec][k]]) then
														if (not SCA.SpellFrame[k]:IsEnabled()) then
															SCA.SpellFrame[k]:Enable()
														end
														SCA.SpellFrame[k]:SetNormalFontObject(SCA.SpellFrame.cFont)
														SCA.SpellFrame[k]:SetHighlightFontObject(SCA.SpellFrame.CFont)
													elseif (k == newSound.validSpell) then
														if (not SCA.SpellFrame[newSound.validSpell]:IsEnabled()) then
															SCA.SpellFrame[newSound.validSpell]:Enable()
														end
														SCA.SpellFrame[newSound.validSpell]:SetNormalFontObject(SCA.SpellFrame.sFont)
														SCA.SpellFrame[newSound.validSpell]:SetHighlightFontObject(SCA.SpellFrame.sFont)
													else
														if (not SCA.SpellFrame[k]:IsEnabled()) then
															SCA.SpellFrame[k]:Enable()
														end
														SCA.SpellFrame[k]:SetNormalFontObject(SCA.SpellFrame.nFont)
														SCA.SpellFrame[k]:SetHighlightFontObject(SCA.SpellFrame.hFont)
													end
												end
											end
											SCA.SpellFrame:Show()
										end
									end,
									width = "half",
								},
								spacing1 = Ace3:Spacer(3,"half"),
								tutorial = Ace3:TutorialToggle(newSound,4,"normal"),
								onCast = Ace3:EventToggle(newSound,'onCast',5,"half","Start","Toggles whether a sound will occur when starting a spell cast.\n\nCombat Log Event: |cFF00FF00SPELL_CAST_START|r","SpecificTab",true),
								onSuccess = Ace3:EventToggle(newSound,'onSuccess',6,"half","Success","Toggles whether a sound will occur when a spell cast completes.\n\nCombat Log Event: |cFF00FF00SPELL_CAST_SUCCESS|r","SpecificTab",true),
								onFail = Ace3:EventToggle(newSound,'onFail',7,"half","Failed","Toggles whether a sound will occur when a spell cast fails.\n\nCombat Log Event: |cFF00FF00SPELL_CAST_FAILED|r","SpecificTab",true),
								onDamage = Ace3:EventToggle(newSound,'onDamage',9,"half","Damage","Toggles whether a sound will occur when starting a spell causes damage.\n\nCombat Log Events: |cFF00FF00SPELL_DAMAGE|r, |cFF00FF00SWING_DAMAGE|r, |cFF00FF00SPELL_PERIODIC_DAMAGE|r","SpecificTab",true),
								onHeal = Ace3:EventToggle(newSound,'onHeal',10,"half","Heal","Toggles whether a sound will occur when starting a spell causes healing.\n\nCombat Log Events: |cFF00FF00SPELL_HEAL|r, |cFF00FF00SPELL_PERIODIC_HEAL|r","SpecificTab",true),
							},
						},
						messageBoard = Ace3:SoundList(newSound,3,true),
						errors = Ace3:Errors("newSound",4,true),
						SoundConfig = {
							name = "Audio Config",
							disabled = function()
								if (newSound.validSpell and newSound.numEvents > 0) then
									return false
								else
									return true
								end
							end,
							hidden = function()
								return not SCA.temp.addSound
							end,
							order = 5,
							type = "group",
							guiInline = true,
							args = {
								source = Ace3:Source(newSound,'newSound',SOUND_SOURCE_OPTIONS,true),
								channel = Ace3:Channel(newSound,SOUND_CHANNEL_OPTIONS,true),
								volume = Ace3:Volume(newSound),
								--[[-----------------------------
									SCA Sound Packs
								-------------------------------]]
								packSelect = Ace3:PackSelect(newSound,"SpecificTab",true),
								--[[-----------------------------
									SharedMedia
								-------------------------------]]
								sharedMedia = Ace3:SharedMedia(newSound,"SpecificTab",true),
								--[[-----------------------------
									Custom Audio
								-------------------------------]]
								audioSource = Ace3:AudioSource(newSound,SOUND_CUSTOM_TYPES,"SpecificTab",true),
								soundEvents = {
									order = 5,
									type = "select",
									name = "Spell Events",
									desc = "Choose whether your custom audio is a file or sound pack",
									disabled = function()
										if (newSound.isSeparate and newSound.validSpell and newSound.numEvents > 1) then
											return false
										else
											return true
										end
									end,
									get = function()
										if (newSound.events.onCast and Ace3:CheckSpell('onCast',newSound.validSpell)) then
											if ((newSound.selected ~= "onCast" and not newSound.events.onSuccess and not newSound.events.onFail and (not newSound.events.onDamage or not Ace3:CheckSpell('onDamage',newSound.validSpell)) and (not newSound.events.onHeal or not Ace3:CheckSpell('onHeal',newSound.validSpell))) or newSound.selected == "empty") then
												newSound.selected = 'onCast'
											end
										else
											if (newSound.selected == "onCast") then
												for k,v in pairs(newSound.events) do
													if (v) then
														newSound.selected = k
														break
													end
												end
												
												if (newSound.selected == "onCast") then
													newSound.selected = "empty"
												end
											end
										end
										
										if (newSound.events.onSuccess and Ace3:CheckSpell('onSuccess',newSound.validSpell)) then
											if ((newSound.selected ~= "onSuccess" and not newSound.events.onCast and not newSound.events.onFail and (not newSound.events.onDamage or not Ace3:CheckSpell('onDamage',newSound.validSpell)) and (not newSound.events.onHeal or not Ace3:CheckSpell('onHeal',newSound.validSpell))) or newSound.selected == "empty") then
												newSound.selected = 'onSuccess'
											end
										else
											if (newSound.selected == "onSuccess") then
												for k,v in pairs(newSound.events) do
													if (v) then
														newSound.selected = k
														break
													end
												end
												
												if (newSound.selected == "onSuccess") then
													newSound.selected = "empty"
												end
											end
										end
										
										if (newSound.events.onFail and Ace3:CheckSpell('onFail',newSound.validSpell)) then
											if ((newSound.selected ~= "onFail" and not newSound.events.onCast and not newSound.events.onSuccess and (not newSound.events.onDamage or not Ace3:CheckSpell('onDamage',newSound.validSpell)) and (not newSound.events.onHeal or not Ace3:CheckSpell('onHeal',newSound.validSpell))) or newSound.selected == "empty") then
												newSound.selected = 'onFail'
											end
										else
											if (newSound.selected == "onFail") then
												for k,v in pairs(newSound.events) do
													if (v) then
														newSound.selected = k
														break
													end
												end
												
												if (newSound.selected == "onFail") then
													newSound.selected = "empty"
												end
											end
										end
										
										if (newSound.events.onDamage and Ace3:CheckSpell('onDamage',newSound.validSpell)) then
											if ((newSound.selected ~= 'onDamage' and not newSound.events.onCast and not newSound.events.onSuccess and not newSound.events.onFail and (not newSound.events.onHeal or not Ace3:CheckSpell('onHeal',newSound.validSpell))) or newSound.selected == "empty") then
												newSound.selected = 'onDamage'
											end
										else
											if (newSound.selected == "onDamage") then
												for k,v in pairs(newSound.events) do
													if (v) then
														newSound.selected = k
														break
													end
												end
												
												if (newSound.selected == "onDamage") then
													newSound.selected = "empty"
												end
											end
										end
										
										if (newSound.events.onHeal and Ace3:CheckSpell('onHeal',newSound.validSpell)) then
											if ((newSound.selected ~= 'onHeal' and not newSound.events.onCast and not newSound.events.onSuccess and not newSound.events.onFail and (not newSound.events.onDamage or not Ace3:CheckSpell('onDamage',newSound.validSpell))) or newSound.selected == "empty") then
												newSound.selected = 'onHeal'
											end
										else
											if (newSound.selected == "onHeal") then
												for k,v in pairs(newSound.events) do
													if (v) then
														newSound.selected = k
														break
													end
												end
												
												if (newSound.selected == "onHeal") then
													newSound.selected = "empty"
												end
											end
										end

										return newSound.selected
									end,
									set = function(self,value)
										newSound.selected = value
										
										newSound.tutorial.flags.spellEvent = true
										
										Ace3:UpdateTutorial(newSound)
									end,
									values = function()
										local args = {}

										for k,v in pairs(newSound.events) do
											if (v and Ace3:CheckSpell(k,newSound.validSpell)) then
												args[k] = SOUND_EVENTS[k]
											end
										end
										
										if (newSound.numEvents == 0) then
											args["empty"] = "No events"
										end
										
										return args
									end
								},
								toggleSeparate = Ace3:ToggleSeparate(newSound,'SpecificTab',true),
								isSCARandom = Ace3:IsSCARandom(newSound,true),
								audioLocation = Ace3:AudioLocation(newSound,true),
								numFiles = Ace3:NumFiles(newSound,'newSound','SpecificTab',true),
								test = Ace3:TestSound(newSound,'newSound','SpecificTab',true),
								isCustomRandom = Ace3:IsCustomRandom(newSound),
								reset = Ace3:SoftReset(newSound,'SpecificTab',true),
							},
						},
						DamageTriggerHeader = {
							order = 6,
							type = 'group',
							name = 'Damage Triggers',
							guiInline = true,
							hidden = function()
								if (SCA.temp.addSound and newSound.events.onDamage and Ace3:CheckSpell('onDamage',newSound.validSpell)) then
									return false
								else
									return true
								end
							end,
							args = {
								threshold = Ace3:Threshold(newSound,'onDamage','Damage','newSound','SpecificTab'),
								critSelect = Ace3:CritSelect(newSound,'onDamage',SOUND_CRIT_TRIGGERS),
							},
						},
						HealingTriggerHeader = {
							order = 7,
							type = 'group',
							name = 'Healing Triggers',
							guiInline = true,
							hidden = function()
								if (SCA.temp.addSound and newSound.events.onHeal and Ace3:CheckSpell('onHeal',newSound.validSpell)) then
									return false
								else
									return true
								end
							end,
							args = {
								threshold = Ace3:Threshold(newSound,'onHeal','Healing','newSound','SpecificTab'),
								critSelect = Ace3:CritSelect(newSound,'onHeal',SOUND_CRIT_TRIGGERS),
							},
						},
						submit = {
							order = 8,
							type = "execute",
							name = function()
								if (SCA.temp.editSound) then
									return "Edit Sound"
								else
									return "Add Sound"
								end
							end,
							disabled = function()
								if (newSound.isSeparate) then
									local numValidEvents = 0
									
									for k,v in pairs(newSound.events) do
										if (v and newSound[k].validAudio and Ace3:CheckSpell(k,newSound.validSpell)) then
											numValidEvents = numValidEvents + 1
										end
									end
									
									if (numValidEvents == newSound.numEvents) then
										return false
									else
										return true
									end
								else
									for k,v in pairs(newSound.events) do
										if (v and newSound[k].validAudio and Ace3:CheckSpell(k,newSound.validSpell)) then
											return false
										end
									end
									
									return true
								end
							end,
							hidden = function()
								return not SCA.temp.addSound
							end,
							func = function(self)
								for k,v in pairs(newSound.events) do
									if (v) then
										if ((k ~= "onCast" and k ~= "onSuccess" and k ~= "onFail" and Ace3:CheckSpell(k,newSound.validSpell)) or (k == "onCast" and Ace3:CheckSpell('onCast',newSound.validSpell)) or (k == "onSuccess" and Ace3:CheckSpell('onSuccess',newSound.validSpell)) or (k == "onFail" and Ace3:CheckSpell('onFail',newSound.validSpell))) then
											local _,spellID,spellName = Ace3:CheckSpell('onSuccess',newSound.validSpell)
											local isOverload = 0
											local _,class = UnitClass("player")
											local spec = GetSpecialization()
											
											for k,v in pairs(SCA.overloadSpells[class][spec]) do
												if (spellName ~= v and strfind(v,spellName)) then
													isOverload = k
													break
												end
											end
											
											if (db.audio.spells.ignored[spellID]) then
												db.audio.spells.ignored[spellID] = nil
											end
											
											if (not specific[spellID]) then
												specific[spellID] = {
													isSeparate = newSound.isSeparate
												}
											end
											
											specific[spellID][k] = {}
											specific[spellID][k].options = {}
											specific[spellID][k].options.source = newSound[k].options.source
											specific[spellID][k].options.channel = newSound[k].options.channel

											if (k == "onDamage") then
												specific[spellID][k].options.trigger = newSound[k].options.trigger
												specific[spellID][k].options.threshold = newSound[k].options.threshold
											end
											
											if (k == "onHeal") then
												specific[spellID][k].options.trigger = newSound[k].options.trigger
												specific[spellID][k].options.threshold = newSound[k].options.threshold
											end
											
											if (newSound[k].options.source == "packs") then
												specific[spellID][k].packs = {}
												specific[spellID][k].packs.value = newSound[k].packs.value
												specific[spellID][k].packs.text = newSound[k].packs.text
												specific[spellID][k].packs.path = newSound[k].packs.path
												specific[spellID][k].packs.max = newSound[k].packs.max
												specific[spellID][k].packs.isRandom = {}
												specific[spellID][k].packs.isRandom.enabled = newSound[k].packs.isRandom.enabled
												specific[spellID][k].packs.isRandom.currentFile = newSound[k].packs.isRandom.currentFile
											elseif (newSound[k].options.source == "shared") then
												specific[spellID][k].shared = {}
												specific[spellID][k].shared.value = newSound[k].shared.value
											elseif (newSound[k].options.source == "custom") then
												specific[spellID][k].custom = {}
												specific[spellID][k].custom.path = newSound[k].custom.path
												specific[spellID][k].custom.max = newSound[k].custom.max
												specific[spellID][k].custom.customType = newSound[k].custom.customType
												specific[spellID][k].custom.isRandom = {}
												specific[spellID][k].custom.isRandom.enabled = newSound[k].custom.isRandom.enabled
												specific[spellID][k].custom.isRandom.currentFile = newSound[k].custom.isRandom.currentFile
											end
											
											if (isOverload > 0) then
												if (not specific[isOverload]) then
													specific[isOverload] = {
														isSeparate = newSound.isSeparate
													}
												end
												
												specific[isOverload][k] = {}
												specific[isOverload][k].options = {}
												specific[isOverload][k].options.source = newSound[k].options.source
												specific[isOverload][k].options.channel = newSound[k].options.channel

												if (k == "onDamage") then
													specific[isOverload][k].options.trigger = newSound[k].options.trigger
													specific[isOverload][k].options.threshold = newSound[k].options.threshold
												end
												
												if (k == "onHeal") then
													specific[isOverload][k].options.trigger = newSound[k].options.trigger
													specific[isOverload][k].options.threshold = newSound[k].options.threshold
												end
												
												if (newSound[k].options.source == "packs") then
													specific[isOverload][k].packs = {}
													specific[isOverload][k].packs.value = newSound[k].packs.value
													specific[isOverload][k].packs.text = newSound[k].packs.text
													specific[isOverload][k].packs.path = newSound[k].packs.path
													specific[isOverload][k].packs.max = newSound[k].packs.max
													specific[isOverload][k].packs.isRandom = {}
													specific[isOverload][k].packs.isRandom.enabled = newSound[k].packs.isRandom.enabled
													specific[isOverload][k].packs.isRandom.currentFile = newSound[k].packs.isRandom.currentFile
												elseif (newSound[k].options.source == "shared") then
													specific[isOverload][k].shared = {}
													specific[isOverload][k].shared.value = newSound[k].shared.value
												elseif (newSound[k].options.source == "custom") then
													specific[isOverload][k].custom = {}
													specific[isOverload][k].custom.path = newSound[k].custom.path
													specific[isOverload][k].custom.max = newSound[k].custom.max
													specific[isOverload][k].custom.customType = newSound[k].custom.customType
													specific[isOverload][k].custom.isRandom = {}
													specific[isOverload][k].custom.isRandom.enabled = newSound[k].custom.isRandom.enabled
													specific[isOverload][k].custom.isRandom.currentFile = newSound[k].custom.isRandom.currentFile
												end
											end
										end
									end
								end
								
								if (SCA.temp.errors.newSound) then
									SCA.temp.errors.newSound = nil
								end
								
								Ace3:ResetDB(newSound,true,true)
								
								self.options.args.SpecificTab.args.messageBoard.args.messages.name = ' '
								ReloadSpellIgnoreList(GetSpellOptions())
								ReloadSpellAssignList(GetSpellOptions(),db.audio.spells)
							end,
						},
						spacing = Ace3:Spacer(10,"half"),
						reset = Ace3:HardReset(newSound,"SpecificTab",true),
						assignedSpells = {
							name = "Assigned Spells",
							order = 1,
							desc = "TESTING",
							hidden = function()
								return SCA.temp.addSound
							end,
							type = "group",
							guiInline = true,
							args = {
							},
						},
					},
				},
				ignoreTab = {
					name = IGNORE,
					order = 4,
					type = "group",
					args = {
						input = {
							name = "Spell Input",
							order = 0,
							type = "group",
							desc = "Click on a checkbox to remove a spell from the ignore list.",
							guiInline = true,
							args = {
								spellInfo = {
									type = "input",
									order = 1,				
									name = "Spell ID / Name",
									desc = "Input spell id, or name, to be ignored by the addon.\n\nSpell names and IDs that you input are checked against a pre-built table of spell ID and names. If there is a spell that is missing that you feel should be added, please message me via the addon page.",
									get = function()
											return newIgnore.validSpell
									end,
									set = function(self,value)
										local spellID,name
										local errors = self.options.args.ignoreTab.args.errors.args.messages
										
										CheckIgnoreSpell(newIgnore,errors,Ace3:CheckSpell('onSuccess',value))
									end,
								},
								spells = {
									order = 2,
									type = "execute",
									name = "Spells",
									desc = "List of spells, and spell IDs, that pertain to your class and spec.\n\n|cFFFF9B00Orange Spells|r are spells that have sounds assigned to them.\n\nA |cFF00FF00Green Spell|r is the spell that is currently selected.\n\n|cFF999999Grey Spells|r are spells that are already being ignored.",
									func = function()
										local _,class = UnitClass("player")
										local spec = GetSpecialization()
										if (SCA.SpellFrame:IsShown()) then
											SCA.SpellFrame:Hide()
										else
											for k,_ in pairs(SCA.spellNames[class][spec]) do
												if (SCA.SpellFrame[k]) then
													if (ignored[SCA.spellNames[class][spec][k]]) then
														SCA.SpellFrame[k]:Disable()
														SCA.SpellFrame[k]:SetDisabledFontObject(SCA.SpellFrame.dFont)
													elseif (specific[SCA.spellNames[class][spec][k]]) then
														if (not SCA.SpellFrame[k]:IsEnabled()) then
															SCA.SpellFrame[k]:Enable()
														end
														SCA.SpellFrame[k]:SetNormalFontObject(SCA.SpellFrame.cFont)
														SCA.SpellFrame[k]:SetHighlightFontObject(SCA.SpellFrame.CFont)
													elseif (k == SCA.temp.newIgnore.validSpell) then
														if (not SCA.SpellFrame[SCA.temp.newIgnore.validSpell]:IsEnabled()) then
															SCA.SpellFrame[SCA.temp.newIgnore.validSpell]:Enable()
														end
														SCA.SpellFrame[SCA.temp.newIgnore.validSpell]:SetNormalFontObject(SCA.SpellFrame.sFont)
														SCA.SpellFrame[SCA.temp.newIgnore.validSpell]:SetHighlightFontObject(SCA.SpellFrame.sFont)
													else
														if (not SCA.SpellFrame[k]:IsEnabled()) then
															SCA.SpellFrame[k]:Enable()
														end
														SCA.SpellFrame[k]:SetNormalFontObject(SCA.SpellFrame.nFont)
														SCA.SpellFrame[k]:SetHighlightFontObject(SCA.SpellFrame.hFont)
													end
												end
											end
											SCA.SpellFrame:Show()
										end
									end,
									width = "half",
								},
								submit = {
									order = 3,
									type = "execute",
									name = "Ignore",
									disabled = function()
										if (#newIgnore.validSpell > 0) then
											return false
										else
											return true
										end
									end,
									func = function(self)
										local _,spellID = Ace3:CheckSpell('onSuccess',newIgnore.validSpell)

										ignored[spellID] = name
										if (specific[spellID]) then
											specific[spellID] = nil
											ReloadSpellAssignList(GetSpellOptions(),db.audio.spells)
										end
										
										self.options.args.ignoreTab.args.errors.args.messages.name = " "
										newIgnore.isValidSpell = false
										newIgnore.validSpell = ''
										ignored[spellID] = true

										if (SCA.temp.errors.ignored) then
											SCA.temp.errors.ignored = nil
										end
								
										ReloadSpellIgnoreList(GetSpellOptions())
									end,
								},
							},
						},
						errors = Ace3:Errors("ignored",4),
						ignoreList = {
							name = "Ignored Spells",
							order = 5,
							type = "group",
							dialogInline = true,
							args = {
							},
						},
					},
				},
			},
		}
	end
	
	return spell_options
end

local defense_options

local function GetDefenseOptions()
	SCA.temp.selectedMenuItem = 3
	
	if (SCA.SpellFrame and SCA.SpellFrame:IsShown()) then
		SCA.SpellFrame:Hide()
	end
	
	if (not defense_options) then
		local db = Ace3.db.char
		local defense = db.audio.defense
		
		defense_options = {
			type = "group",
			childGroups = "tab",
			order = 1,
			name = "",
			desc = '',
			args = {
				disabledError = Ace3:DisabledError(defense,true),
				subevent = {
					order = 1,
					type = "select",
					name = "Subevent Select",
					desc = "Choose which event you want to edit",
					get = function()
						return defense.selected
					end,
					set = function(self,value)
						defense.selected = value
					end,
					values = function()
						local subevents = {}

						for k,_ in pairs (defense.events) do
							subevents[k] = defense[k].label
						end
						
						return subevents
					end,
				},
				reset = Ace3:SoftReset(defense,nil,nil,"partial"),
				messageBoard = Ace3:SoundList(defense,13,nil,true),
				errors = Ace3:Errors('defense',14),
				SoundConfig = {
					name = "Audio Config",
					order = 15,
					type = "group",
					guiInline = true,
					args = {
						source = Ace3:Source(defense,'defense',SOUND_SOURCE_OPTIONS,nil,true,true),
						channel = Ace3:Channel(defense,SOUND_CHANNEL_OPTIONS,nil,true),
						volume = Ace3:Volume(defense),
						--[[-----------------------------
							SCA Sound Packs
						-------------------------------]]
						packSelect = Ace3:PackSelect(defense,nil,nil,"partial"),
						isSCARandom = Ace3:IsSCARandom(defense,nil,true),
						--[[-----------------------------
							SharedMedia
						-------------------------------]]
						sharedMedia = Ace3:SharedMedia(defense,nil,nil,"partial"),
						--[[-----------------------------
							Custom Audio
						-------------------------------]]
						audioSource = Ace3:AudioSource(defense,SOUND_CUSTOM_TYPES,nil,nil,true,true),
						isCustomRandom = Ace3:IsCustomRandom(defense,true),
						spacing1 = Ace3:Spacer(6,"half"),
						audioLocation = Ace3:AudioLocation(defense,nil,true,true),
						numFiles = Ace3:NumFiles(defense,'defense',nil,nil,true,true),
						test = Ace3:TestSound(defense,'defense',nil,nil,"partial",true),
					},
				},	
			},
		}
		
	end
	
	return defense_options
end

local function UpdateSpecialMessage(db,options)
	local msg = ''
	local ctr = 1
	
	for k,_ in pairs(db.events) do
		if (ctr > 1) then
			msg = msg.."\n"
		end
		if (db[k].validAudio) then
			msg = msg..GetSoundInfo(db,k,db[k].options.source,db[k].label)
		else
			msg = msg..db[k].label..": |cFFFF5F99Not yet configured!|r"
		end
		ctr = ctr + 1
	end
	options.args.messageBoard.args.messages.name = msg
end

local function SpecialTabBuilder()
	local tabs = {}
	local db = Ace3.db.char

	local special = db.audio.special
	local specTemp = SCA.temp.special

	tabs["disabledError"] = Ace3:DisabledError(special)
	for k,_ in pairs(special) do
		if (type(special[k]) == "table") then
			tabs[k.."Tab"] = {
				type = "group",
				order = special[k].order,
				name = special[k].label,
				desc = '',
				args = {
					disabledError = Ace3:DisabledError(special[k]),
					subevent = {
						order = 1,
						type = "select",
						name = "Subevent Select",
						desc = "Choose which event you want to edit",
						get = function()
							return special[k].selected
						end,
						set = function(self,value)
							special[k].selected = value
						end,
						values = function()
							local subevents = {}

							for l,_ in pairs (special[k].events) do
								subevents[l] = special[k][l].label
							end
							
							return subevents
						end,
					},
					reset = Ace3:SoftReset(special[k],k..'Tab',nil,"partial"),
					messageBoard = Ace3:SoundList(special[k],13,nil,true),
					errors = Ace3:Errors(k,14),
					SoundConfig = {
						name = "Audio Config",
						order = 15,
						type = "group",
						guiInline = true,
						args = {
							source = Ace3:Source(special[k],k,SOUND_SOURCE_OPTIONS,nil,true,true),
							channel = Ace3:Channel(special[k],SOUND_CHANNEL_OPTIONS,nil,true),
							volume = Ace3:Volume(special[k]),
							--[[-----------------------------
								SCA Sound Packs
							-------------------------------]]
							packSelect = Ace3:PackSelect(special[k],k.."Tab",nil,"partial"),
							isSCARandom = Ace3:IsSCARandom(special[k],nil,true),
							--[[-----------------------------
								SharedMedia
							-------------------------------]]
							sharedMedia = Ace3:SharedMedia(special[k],k.."Tab",nil,"partial"),
							--[[-----------------------------
								Custom Audio
							-------------------------------]]
							audioSource = Ace3:AudioSource(special[k],SOUND_CUSTOM_TYPES,k.."Tab",nil,true,true),
							isCustomRandom = Ace3:IsCustomRandom(special[k],true),
							spacing1 = Ace3:Spacer(6,"half"),
							audioLocation = Ace3:AudioLocation(special[k],nil,true,true),
							numFiles = Ace3:NumFiles(special[k],k,k..'Tab',nil,true,true),
							test = Ace3:TestSound(special[k],k,k..'Tab',nil,"partial",true),
						},
					},							
				},
			}
		end
	end

	return tabs
end
			
local special_options

local function GetSpecialOptions()
	SCA.temp.selectedMenuItem = 4
	
	if (SCA.SpellFrame and SCA.SpellFrame:IsShown()) then
		SCA.SpellFrame:Hide()
	end
	
	if (not special_options) then
		special_options = {
			type = "group",
			childGroups = "tab",
			order = 1,
			name = "",
			desc = '',
			args = SpecialTabBuilder(),
		}
	end
	
	return special_options
end

local function ChatTabBuilder()
	local tabs = {}
	local db = Ace3.db.char

	local chat = db.audio.chat

	tabs["disabledError"] = Ace3:DisabledError(chat)
	for k,_ in pairs(chat) do
		if (type(chat[k]) == "table") then
			tabs[k.."Tab"] = {
				type = "group",
				order = chat[k].order,
				name = chat[k].label,
				desc = '',
				args = {
					disabledError = Ace3:DisabledError(chat[k]),
					subevent = {
						order = 1,
						type = "select",
						name = "Subevent Select",
						desc = "Choose which event you want to edit",
						get = function()
							return chat[k].selected
						end,
						set = function(self,value)
							chat[k].selected = value
						end,
						values = function()
							local subevents = {}

							for l,_ in pairs (chat[k].events) do
								subevents[l] = chat[k][l].label
							end
							
							return subevents
						end,
					},
					spacing = Ace3:Spacer(10,"half"),
					reset = Ace3:SoftReset(chat[k],k..'Tab',nil,"partial"),
					messageBoard = Ace3:SoundList(chat[k],13,nil,true),
					errors = Ace3:Errors(k,14),
					SoundConfig = {
						name = "Audio Config",
						order = 15,
						type = "group",
						guiInline = true,
						args = {
							source = Ace3:Source(chat[k],k,SOUND_SOURCE_OPTIONS,nil,true,true),
							channel = Ace3:Channel(chat[k],SOUND_CHANNEL_OPTIONS,nil,true),
							volume = Ace3:Volume(chat[k]),
							--[[-----------------------------
								SCA Sound Packs
							-------------------------------]]
							packSelect = Ace3:PackSelect(chat[k],k.."Tab",nil,"partial"),
							isSCARandom = Ace3:IsSCARandom(chat[k],nil,true),
							--[[-----------------------------
								SharedMedia
							-------------------------------]]
							sharedMedia = Ace3:SharedMedia(chat[k],k.."Tab",nil,"partial"),
							--[[-----------------------------
								Custom Audio
							-------------------------------]]
							audioSource = Ace3:AudioSource(chat[k],SOUND_CUSTOM_TYPES,k.."Tab",nil,true,true),
							isCustomRandom = Ace3:IsCustomRandom(chat[k],true),
							spacing1 = Ace3:Spacer(6,"half"),
							audioLocation = Ace3:AudioLocation(chat[k],nil,true,true),
							numFiles = Ace3:NumFiles(chat[k],k,k..'Tab',nil,true,true),
							test = Ace3:TestSound(chat[k],k,k..'Tab',nil,"partial",true),
						},
					},
					ChatConfig = {
						name = "Chat Config",
						order = 16,
						type = "group",
						guiInline = true,
						args = {
							toggle = {
								order = 1,
								type = "toggle",
								name = "Timeout",
								desc = "Toggles a timeout that will delay sounds from playing based on the provided delay amount",
								get = function()
									return chat[k][chat[k].selected].options.timeout.enabled
								end,
								set = function(self,value)
									chat[k][chat[k].selected].options.timeout.enabled = value
								end,
							},
							delay = {
								order = 2,
								type = "input",
								name = "Delay",
								desc = "Set a delay value, in seconds.",
								disabled = function()
									if (chat[k][chat[k].selected].options.timeout.enabled) then
										return false
									else
										return true
									end
								end,
								get = function()
									return chat[k][chat[k].selected].options.timeout.delay
								end,
								set = function(self,value)
									chat[k][chat[k].selected].options.timeout.delay = value
								end,
								width = "half",
							},
							ignoreMe = {
								order = 3,
								type = "toggle",
								name = "Ignore Me",
								desc = "Do not play a sound if that chat message was sent by you",
								get = function()
									return chat[k][chat[k].selected].options.ignoreMe
								end,
								set = function(self,value)
									chat[k][chat[k].selected].options.ignoreMe = value
								end,
							},
						},
					},
				},
			}
		end
	end

	return tabs
end

local chat_options

local function GetChatOptions()
	SCA.temp.selectedMenuItem = 5
	
	if (SCA.SpellFrame and SCA.SpellFrame:IsShown()) then
		SCA.SpellFrame:Hide()
	end
	
	if (not chat_options) then
		chat_options = {
			type = "group",
			childGroups = "tab",
			order = 1,
			name = "",
			desc = '',
			args = ChatTabBuilder(),
		}
	end
	
	return chat_options
end

local function MoveTabBuilder()
	local tabs = {}
	local db = Ace3.db.char

	local move = db.audio.move

	tabs["disabledError"] = Ace3:DisabledError(move)
	for k,_ in pairs(move) do
		if (type(move[k]) == "table") then
			tabs[k.."Tab"] = {
				type = "group",
				order = move[k].order,
				name = move[k].label,
				desc = function()
					if (k == "forward") then
						return "|cFFFF0000Does not support moving forward with the mouse|r"
					end
				end,
				args = {
					disabledError = Ace3:DisabledError(move[k]),
					options = {
						name = "General Options",
						order = 1,
						type = "group",
						guiInline = true,
						args = {
							onStart = Ace3:EventToggle(move[k],'onStart',1,"normal","On Start","Toggle whether or not a sound will play when starting to "..lower(move[k].label),k.."Tab",nil,nil,true),
							onStop = Ace3:EventToggle(move[k],'onStop',2,"normal","On Stop","Toggle whether or not a sound will play when stopping "..lower(move[k].label),k.."Tab",nil,nil,true),
						},
					},
					triggers = {
						name = "Sound Triggers",
						order = 2,
						hidden = function()
							if (k == "pitchup" or k == "pitchdown" or k == "autorun") then
								return true
							else
								return false
							end
						end,
						type = "group",
						guiInline = true,
						args = {
							running = {
								order = 1,
								type = "toggle",
								name = "On Foot",
								desc = "Toggle whether or not a sound will play while on foot",
								get = function()
									return move[k].triggers.running
								end,
								set = function(self,value)
									move[k].triggers.running = value
								end,
								width = "half",
							},
							mounted = {
								order = 2,
								type = "toggle",
								name = "Mounted",
								desc = "Toggle whether or not a sound will play while mounted",
								get = function()
									return move[k].triggers.mounted
								end,
								set = function(self,value)
									move[k].triggers.mounted = value
								end,
								width = "half",
							},
							flying = {
								order = 3,
								type = "toggle",
								name = "Flying",
								disabled = function()
									if (move[k].triggers.mounted) then
										return false
									else
										return true
									end
								end,
								desc = "Toggle whether or not a sound will play while flying",
								get = function()
									return move[k].triggers.flying
								end,
								set = function(self,value)
									move[k].triggers.flying = value
								end,
								width = "half",
							},
							swimming = {
								order = 4,
								type = "toggle",
								name = "Swimming",
								desc = "Toggle whether or not a sound will play while swimming",
								get = function()
									return move[k].triggers.swimming
								end,
								set = function(self,value)
									move[k].triggers.swimming = value
								end,
								width = "normal",
							},
							
						},
					},
					messageBoard = Ace3:SoundList(move[k],3),
					errors = Ace3:Errors(k,4),
					SoundConfig = {
						name = "Audio Config",
						order = 5,
						type = "group",
						guiInline = true,
						args = {
							source = Ace3:Source(move[k],k,SOUND_SOURCE_OPTIONS,nil,nil,true),
							channel = Ace3:Channel(move[k],SOUND_CHANNEL_OPTIONS),
							volume = Ace3:Volume(move[k]),
							--[[-----------------------------
								--SCA Sound Packs
							-------------------------------]]
							packSelect = Ace3:PackSelect(move[k],k.."Tab"),
							isSCARandom = Ace3:IsSCARandom(move[k]),
							--[[-----------------------------
								--SharedMedia
							-------------------------------]]
							sharedMedia = Ace3:SharedMedia(move[k],k.."Tab"),
							--[[-----------------------------
								--Custom Audio
							-------------------------------]]
							audioSource = Ace3:AudioSource(move[k],SOUND_CUSTOM_TYPES,k.."Tab",nil,nil,true),
							actions = {
								order = 5,
								type = "select",
								name = "Actions",
								desc = "Choose whether your custom audio is a file or sound pack",
								disabled = function()
									if (move[k].isSeparate and move[k].numEvents > 1) then
										return false
									else
										return true
									end
								end,
								get = function()
									if (move[k].events.onStart) then
										if (move[k].selected ~= "onStart" and not move[k].events.onStop) then
											move[k].selected = 'onStart'
										end
									else
										if (move[k].selected == "onStart") then
											for j,v in pairs(move[k].events) do
												if (v) then
													move[k].selected = j
													break
												end
											end
														
											if (move[k].selected == "onStart") then
												move[k].selected = "empty"
											end
										end
									end
									
									if (move[k].events.onStop) then
										if (move[k].selected ~= "onStop" and not move[k].events.onStart) then
											move[k].selected = 'onStop'
										end
									else
										if (move[k].selected == "onStop") then
											for j,v in pairs(move[k].events) do
												if (v) then
													move[k].selected = j
													break
												end
											end
														
											if (move[k].selected == "onStop") then
												move[k].selected = "empty"
											end
										end
									end
									
									return move[k].selected
								end,
								set = function(self,value)
									move[k].selected = value
								end,
								values = function()
									local args,numEvents = {},0
									
									for j,v in pairs(move[k].events) do
										if (v) then
											numEvents = numEvents + 1
											args[j] = MOVEMENT_ACTIONS[j]
										end
									end
									
									if (numEvents == 0) then
										args["empty"] = "No actions"
									end
									
									return args
								end
							},
							toggleSeparate = Ace3:ToggleSeparate(move[k],k..'Tab',nil,true),
							isSCARandom = Ace3:IsSCARandom(move[k]),
							audioLocation = Ace3:AudioLocation(move[k],nil,nil,true),
							numFiles = Ace3:NumFiles(move[k],k,k..'Tab',nil,nil,true),
							test = Ace3:TestSound(move[k],k,k..'Tab',nil,nil,true),
							isCustomRandom = Ace3:IsCustomRandom(move[k]),
							reset = Ace3:SoftReset(move[k],k..'Tab'),
						},
					},
				},
			}
		end
	end

	return tabs
end

local move_options

local function GetMoveOptions()
	SCA.temp.selectedMenuItem = 6
	
	if (SCA.SpellFrame and SCA.SpellFrame:IsShown()) then
		SCA.SpellFrame:Hide()
	end
	
	if (not move_options) then
		move_options = {
			type = "group",
			childGroups = "tab",
			order = 1,
			name = "",
			desc = '',
			args = MoveTabBuilder(),
		}
	end
	
	return move_options
end

function Ace3:SetupOptions()
	local ACFG = LibStub("AceConfig-3.0")
	ACFG:RegisterOptionsTable("CustomAudio Quick Toggle", GetToggleOptions)
	ACFG:RegisterOptionsTable("CustomAudio Spell Sounds", GetSpellOptions)
	ACFG:RegisterOptionsTable("CustomAudio Defense Sounds", GetDefenseOptions)
	ACFG:RegisterOptionsTable("CustomAudio Special Sounds", GetSpecialOptions)
	ACFG:RegisterOptionsTable("CustomAudio Chat Sounds", GetChatOptions)
	ACFG:RegisterOptionsTable("CustomAudio Movement Sounds", GetMoveOptions)

	local ACD = LibStub("AceConfigDialog-3.0")
	ACD:AddToBlizOptions("CustomAudio Quick Toggle", "Quick Toggle", "CustomAudio")
	ACD:AddToBlizOptions("CustomAudio Spell Sounds", "Spell Sounds", "CustomAudio")
	ACD:AddToBlizOptions("CustomAudio Defense Sounds", "Defense Sounds", "CustomAudio")
	ACD:AddToBlizOptions("CustomAudio Special Sounds", "Special Sounds", "CustomAudio")
	ACD:AddToBlizOptions("CustomAudio Chat Sounds", "Chat Sounds", "CustomAudio")
	ACD:AddToBlizOptions("CustomAudio Movement Sounds", "Movement Sounds", "CustomAudio")
end

local function CreateSpellFrame()
	local sortTable = {}
	local class,EnglishClass = UnitClass("player")
	local _,specName = GetSpecializationInfo(GetSpecialization())
	local spec = GetSpecialization()
	local SpellFrame = CreateFrame("frame")
	
	SpellFrame:SetWidth(200)
	SpellFrame:SetPoint("TOPLEFT",InterfaceOptionsFrame,"TOPLEFT",355,-25)
	SpellFrame:SetFrameStrata("TOOLTIP")
	SpellFrame:SetBackdrop(SCA.BackdropSB)
	SpellFrame:SetBackdropColor(0,0,0,1)
	SpellFrame:SetBackdropBorderColor(1,1,1,1)
	SpellFrame:Hide()
	
	SpellFrame.text = SpellFrame:CreateFontString(nil,'HIGH', 'GameFontHighlightLarge')
	SpellFrame.text:SetWidth(SpellFrame:GetWidth())
	SpellFrame.text:SetHeight(20)
	SpellFrame.text:SetPoint("TOPLEFT",SpellFrame,"TOPLEFT",0,0)
	SpellFrame.text:SetFont("Fonts\\FRIZQT__.TTF", 8)
	SpellFrame.text:SetTextColor(0,1,0,1)
	SpellFrame.text:SetText(specName.." "..class.." Spells")
	
	SpellFrame.dFont = CreateFont("SCA Disabled Font")
	SpellFrame.dFont:SetJustifyH("LEFT")
	SpellFrame.dFont:SetTextColor(0.3,0.3,0.3,1)
	SpellFrame.dFont:SetFont("Fonts\\FRIZQT__.TTF", 8)
	
	SpellFrame.eFont = CreateFont("SCA Error Font")
	SpellFrame.eFont:SetJustifyH("LEFT")
	SpellFrame.eFont:SetTextColor(1,0,0,1)
	SpellFrame.eFont:SetFont("Fonts\\FRIZQT__.TTF", 8)
	
	SpellFrame.cFont = CreateFont("SCA Caution Font")
	SpellFrame.cFont:SetJustifyH("LEFT")
	SpellFrame.cFont:SetTextColor(1,0.6,0,1)
	SpellFrame.cFont:SetFont("Fonts\\FRIZQT__.TTF", 8)
	
	SpellFrame.CFont = CreateFont("SCA Caution Highlight Font")
	SpellFrame.CFont:SetJustifyH("LEFT")
	SpellFrame.CFont:SetTextColor(1,0.6,0,1)
	SpellFrame.CFont:SetFont("Fonts\\FRIZQT__.TTF", 9)
	
	SpellFrame.nFont = CreateFont("SCA Normal Font")
	SpellFrame.nFont:SetJustifyH("LEFT")
	SpellFrame.nFont:SetTextColor(1,1,1,1)
	SpellFrame.nFont:SetFont("Fonts\\FRIZQT__.TTF", 8)
	
	SpellFrame.hFont = CreateFont("SCA Highlight Font")
	SpellFrame.hFont:SetJustifyH("LEFT")
	SpellFrame.hFont:SetTextColor(1,1,0.25,1)
	SpellFrame.hFont:SetFont("Fonts\\FRIZQT__.TTF", 9)
	
	SpellFrame.sFont = CreateFont("SCA Selected Font")
	SpellFrame.sFont:SetJustifyH("LEFT")
	SpellFrame.sFont:SetTextColor(0.1,1,0.25,1)
	SpellFrame.sFont:SetFont("Fonts\\FRIZQT__.TTF", 9)
	
	for k,v in pairs(SCA.spellNames[EnglishClass][spec]) do
		tinsert(sortTable,k.." ("..v..")")
	end
	
	SpellFrame:SetHeight((#sortTable * 10) + 45)
	
	table.sort(sortTable)
	
	for i=1,#sortTable do	
		local spellNameLoc = string.find(sortTable[i],"%(")
		local spellName = sub(sortTable[i],0,(spellNameLoc-2))
		
		SpellFrame[spellName] = CreateFrame("Button",nil,SpellFrame)
		SpellFrame[spellName]:SetWidth(SpellFrame:GetWidth())
		SpellFrame[spellName]:SetHeight(20)
		if (i == 1) then
			SpellFrame[spellName]:SetPoint("TOPLEFT",SpellFrame,"TOPLEFT",0,-20)
		else
			SpellFrame[spellName]:SetPoint("TOPLEFT",SpellFrame,"TOPLEFT",0,(-10*i)-10)
		end
		
		if (SCA.pvpSpells[EnglishClass][spec][SCA.spellNames[EnglishClass][spec][spellName]]) then
			SpellFrame[spellName]:SetText(sortTable[i].." |cFFFF0000(PvP)|r")
		elseif (SCA.artifactSpells[EnglishClass][spec][SCA.spellNames[EnglishClass][spec][spellName]]) then
			SpellFrame[spellName]:SetText(sortTable[i].." |cFF00F3FF(Artifact)|r")
		else
			SpellFrame[spellName]:SetText(sortTable[i])
		end
		
		SpellFrame[spellName]:SetNormalFontObject(SpellFrame.nFont)
		SpellFrame[spellName]:SetHighlightFontObject(SpellFrame.hFont)
		SpellFrame[spellName]:Show()

		SpellFrame[spellName]:SetScript("OnClick",function(self,button)
			local spellNameLoc = string.find(self:GetText(),"%(")
			local spellName = sub(self:GetText(),0,(spellNameLoc-2))
			local newItem
			
			if (SCA.temp.addSound) then
				local errors = GetSpellOptions().args.SpecificTab.args.errors.args.messages
				newItem = SCA.temp.newSound
				newItem.validSpell = spellName
				
				CheckSpecificSpell(newItem,errors,Ace3:CheckSpell('onSuccess',spellName))
				Ace3:UpdateTutorial(newItem)
			else
				local errors = GetSpellOptions().args.ignoreTab.args.errors.args.messages
				newItem = SCA.temp.newIgnore
				newItem.validSpell = spellName

				CheckIgnoreSpell(newItem,errors,Ace3:CheckSpell('onSuccess',spellName))
			end
			
			Ace3:NavigateInterfaceOptions(SCA.temp.selectedMenuItem,true)

			SpellFrame:Hide()
		end)
	end

	SpellFrame.close = CreateFrame("Button",nil,SpellFrame)
	SpellFrame.close:SetWidth(SpellFrame:GetWidth())
	SpellFrame.close:SetHeight(20)
	SpellFrame.close:SetPoint("TOPLEFT",SpellFrame,"TOPLEFT",0,floor((SpellFrame:GetHeight() - 20) * -1))
	SpellFrame.close:SetText("Close")
	SpellFrame.close:SetNormalFontObject(SpellFrame.nFont)
	SpellFrame.close:SetHighlightFontObject(SpellFrame.hFont)
	SpellFrame.close:Show()

	SpellFrame.close:SetScript("OnClick",function(self,button)
		SpellFrame:Hide()
	end)
		
	SCA.SpellFrame = SpellFrame
end

local function UpdateInterfaceSettings()
	local db = Ace3.db.char
	
	InitSpellOptions(db.audio.spells.damage,GetSpellOptions().args.DamageTab)
	InitSpellOptions(db.audio.spells.heals,GetSpellOptions().args.HealingTab)
	InitMoveOptions(db.audio.move)

	UpdateSpecialMessage(db.audio.defense,GetDefenseOptions())
	
	for k,_ in pairs(db.audio.special) do
		if (type(db.audio.special[k]) == "table") then
			UpdateSpecialMessage(db.audio.special[k],GetSpecialOptions().args[k.."Tab"])
		end
	end
	
	for k,_ in pairs(db.audio.chat) do
		if (type(db.audio.chat[k]) == "table") then
			UpdateSpecialMessage(db.audio.chat[k],GetChatOptions().args[k.."Tab"])
		end
	end
	
	for k,_ in pairs(db.audio.move) do
		if (type(db.audio.move[k]) == "table") then
			UpdateSpecialMessage(db.audio.move[k],GetMoveOptions().args[k.."Tab"])
		end
	end
	
	ReloadSpellIgnoreList(GetSpellOptions())
	ReloadSpellAssignList(GetSpellOptions(),db.audio.spells)
	
	CreateSpellFrame()
end

InterfaceOptionsFrame:HookScript("OnUpdate",function(self)
	-- The purpose here is to refresh the contents of the addon's config panel; simulating a click to the left navigation items will do this.
	-- This will only be done if the SHIFT key is down (making sure to check if it's down only once, via the "isShiftDown" variable).
	-- We also want to make sure the keyboard doesn't have focus, as SHIFT is commonly pressed while chatting.
	if (IsShiftKeyDown() and not SCA.temp.isShiftDown and not GetCurrentKeyBoardFocus()) then
		Ace3:NavigateInterfaceOptions(SCA.temp.selectedMenuItem,true)
		SCA.temp.isShiftDown = true
	elseif (not IsShiftKeyDown() and SCA.temp.isShiftDown) then
		Ace3:NavigateInterfaceOptions(SCA.temp.selectedMenuItem,true)
		SCA.temp.isShiftDown = false
	end
end)

InterfaceOptionsFrame:HookScript("OnShow",function(self)
	UpdateInterfaceSettings()
end)

InterfaceOptionsFrame:HookScript("OnHide",function(self)
	if (SCA.SpellFrame:IsShown()) then
		SCA.SpellFrame:Hide()
	end
end)